# Copyright 2025 Cardy Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pytest import approx
from cardy import distance, max_distance, norm_distance
from utils import test


@test
def empty_card_sorts_have_a_distance_of_zero():
    assert distance((), ()) == 0


@test
def equivalent_card_sorts_have_an_edit_distance_of_zero():
    assert distance(({1},), ({1},)) == 0
    assert distance(({1}, {2}), ({1}, {2})) == 0
    assert distance(({2}, {1}), ({1}, {2})) == 0
    assert distance(
        ({1, 2, 3}, {4}, {5, 6}),
        ({4}, {1, 2, 3}, {5, 6}),
    ) == 0


@test
def empty_groups_are_ignored_when_computing_distances():
    assert distance(({1}, {2}, set()), ({1}, {2})) == 0
    assert distance(({1}, {2}), ({1}, set(), {2})) == 0


@test
def single_card_displacements_have_an_edit_distance_of_one():
    assert distance(({1, 2}, {3}), ({1}, {2, 3})) == 1
    assert distance(({1}, {2}, {3}), ({1}, {2, 3})) == 1
    assert distance(({1}, {2, 3}), ({1, 2, 3},)) == 1


@test
def distance_between_card_sorts_is_computed_for_multiple_moves():
    assert distance(
        ({1, 2, 3}, {4, 5, 6}, {7, 8, 9}),
        ({1, 2}, {3, 4}, {5, 6, 7}, {8, 9}),
    ) == 3
    assert distance(
        ({1, 2}, {3, 4}, {5, 6, 7}, {8, 9}),
        ({1, 2, 3}, {4, 5, 6}, {7, 8, 9}),
    ) == 3


@test
def maximum_edit_distance_is_computed_for_card_sorts():
    sort = (
        {"a1", "a2", "a3"},
        {"b1", "b2", "b3", "b4", "b5"},
        {"c1", "c2", "c3", "c4", "c5"},
        {"d1", "d2", "d3", "d4"},
    )
    assert max_distance(sort) == 13
    assert max_distance(sort, num_groups=4) == 12


@test
def normalised_edit_distance_is_computed_for_card_sorts():
    sort1 = (
        {"a1", "a2", "a3"},
        {"b1", "b2", "b3", "b4", "b5"},
        {"c1", "c2", "c3", "c4", "c5"},
        {"d1", "d2", "d3", "d4"},
    )
    sort2 = (
        {"a1", "b1", "b5", "c1", "c5", "d1"},
        {"a2", "b2", "c2", "d2"},
        {"a3", "b3", "c3", "d3"},
        {"b4", "c4", "d4"},
    )
    sort3 = (
        {"a1", "b1", "c1", "d1"},
        {"a2", "b2", "c2", "d2"},
        {"a3", "b3", "c3", "d3"},
        {"b4", "c4", "d4"},
        {"b5", "c5"},
    )
    sort4 = (
        {"a1", "a2"},
        {"a3", "b1", "b2", "b3", "b4", "b5"},
        {"c1", "c2", "c3", "c4", "c5"},
        {"d1", "d2", "d3", "d4"},
    )
    assert norm_distance(sort1, sort2, num_groups=4) == 1
    assert norm_distance(sort1, sort3) == 1
    assert norm_distance(sort1, sort4, num_groups=4) == approx(1 / 12)



# Copyright 2025 Cardy Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cardy import neighbourhood, norm_distance
from utils import test

SORTS = {
    0: ({1, 2, 3}, {4, 5}),
    1: ({1, 2, 3}, {4, 5}, set()),
    2: ({1, 2}, {3}, {4, 5}),
    3: ({1, 2}, {3, 4}, {5}),
    4: ({1, 2, 4}, {3, 5},),
}


@test
def zero_neighbourhoods_only_include_equivalent_card_sorts():
    assert neighbourhood(0, SORTS[0], SORTS) == {0, 1}
    assert neighbourhood(0, SORTS[2], SORTS) == {2}
    assert neighbourhood(0, ({1, 2, 3, 4, 5},), SORTS) == set()


@test
def zero_norm_neighbourhoods_only_include_equivalent_card_sorts():
    dist = lambda l, r: norm_distance(l, r, num_groups=3)
    assert neighbourhood(0, SORTS[0], SORTS, distance=dist) == {0, 1}
    assert neighbourhood(0, SORTS[2], SORTS, distance=dist) == {2}
    assert neighbourhood(0, ({1, 2, 3, 4, 5},), SORTS, distance=dist) == set()


@test
def neighbourhoods_are_returned_when_sorts_have_different_distances():
    assert neighbourhood(1, SORTS[0], SORTS) == {0, 1, 2}
    assert neighbourhood(2, SORTS[0], SORTS) == {0, 1, 2, 3, 4}
    assert neighbourhood(2, ({1, 2, 3, 4, 5},), SORTS) == {0, 1, 4}
    assert neighbourhood(3, ({1, 2, 3, 4, 5},), SORTS) == {0, 1, 2, 3, 4}


@test
def neighbourhoods_are_returned_when_sorts_have_different_norm_distances():
    dist = lambda l, r: norm_distance(l, r, num_groups=3)
    assert neighbourhood(1 / 3, SORTS[0], SORTS, distance=dist) \
           == {0, 1, 2}
    assert neighbourhood(2 / 3, SORTS[0], SORTS, distance=dist) \
           == {0, 1, 2, 3, 4}
    assert neighbourhood(2 / 3, ({1, 2, 3, 4, 5},), SORTS, distance=dist) \
           == {0, 1, 4}
    assert neighbourhood(3 / 3, ({1, 2, 3, 4, 5},), SORTS, distance=dist) \
           == {0, 1, 2, 3, 4}


# Copyright 2025 Cardy Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pytest import approx

from cardy import orthogonality, norm_distance
from cardy.orthogonality import DisjointSet, min_spanning_tree
from utils import test

EXAMPLE_2 = (
    ({9, 17, 18, 25},
     {1, 3, 5, 8, 12, 13, 20, 26},
     {2, 6, 7, 10, 11, 15, 19, 22},
     {4, 14, 16, 21, 23},
     {24}),
    ({12, 20},
     {18, 25},
     {3, 13, 19},
     {5, 10, 15},
     {1, 17},
     {8, 9},
     {2, 4, 6, 7, 11, 14, 16, 21, 22, 23, 24, 26}),
    ({1, 3, 5, 6, 7, 9, 10, 11, 12, 14, 15, 16, 18, 19, 20, 21, 25, 26},
     {2, 4, 8, 13, 17, 22, 23, 24})
)

EXAMPLE_1 = (
    ({1, 3, 4, 5, 6, 7, 13, 14, 15, 22, 23},
     {2, 8, 9, 10, 11, 12, 16, 17, 18, 19, 20, 21, 24, 25, 26}),
    ({1, 3, 4, 6, 7, 10, 13, 14, 15, 18, 23, 26},
     {2, 5, 8, 9, 11, 12, 16, 17, 19, 20, 21, 22, 24, 25}),
    ({1, 2, 5, 8, 9, 11, 12, 16, 17, 18, 19, 20, 21, 22, 24, 25},
     {3, 4, 6, 7, 10, 13, 14, 15, 23, 26}),
)


@test
def disjoint_sets():
    forest = DisjointSet(range(1, 11))
    for i in range(1, 11):
        assert forest.find(i) == i
    forest.merge(1, 2)
    forest.merge(2, 3)
    forest.merge(3, 3)
    forest.merge(3, 4)
    root = forest.find(1)
    forest.merge(5, 6)
    forest.merge(6, 4)
    assert forest.find(6) == root


@test
def spanning_trees():
    vertices = ["A", "B", "C", "D", "E", "F", "G"]
    edges = {
        ('A', 'B'): 7,
        ('A', 'D'): 5,
        ('B', 'C'): 8,
        ('B', 'D'): 9,
        ('B', 'E'): 7,
        ('C', 'E'): 5,
        ('D', 'E'): 15,
        ('D', 'F'): 6,
        ('E', 'F'): 8,
        ('E', 'G'): 9,
        ('F', 'G'): 11,
    }
    assert min_spanning_tree(vertices, edges) == {
        ('E', 'G'), ('A', 'B'), ('D', 'F'), ('C', 'E'), ('B', 'E'), ('A', 'D')
    }


@test
def nmst_metric():
    # Examples from: https://doi.org/10.1111/j.1468-0394.2005.00305.x
    dist1 = lambda l, r: norm_distance(l, r, num_groups=2)
    assert orthogonality(EXAMPLE_1, distance=dist1) == approx(7 / 39)
    dist2 = lambda l, r: norm_distance(l, r, num_groups=7)
    assert orthogonality(EXAMPLE_2, distance=dist2) == approx(11 / 21)



# Copyright 2025 Cardy Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from munkres import Munkres, make_cost_matrix

from .types import CardSort

__all__ = ("distance", "max_distance", "norm_distance")


def distance[T](sort1: CardSort[T], sort2: CardSort[T]) -> int:
    """Computes the edit distance between the two given card sorts."""
    if not sort1 and not sort2:
        return 0

    weights = [
        [len(group1 & group2) for group2 in sort2]
        for group1 in sort1
    ]
    cost_matrix = make_cost_matrix(weights)
    total = sum([
        weights[row][col]
        for row, col in Munkres().compute(cost_matrix)
    ])
    return sum(len(g) for g in sort1) - total


def max_distance[T](sort: CardSort[T], *, num_groups: int = None) -> int:
    """
    Computes the maximum edit distance any other card sort could be to the
    given card sort
    """
    if num_groups is None:
        num_groups = max(len(g) for g in sort)
    weights = []
    for group in sort:
        k = len(group) - len(group) // num_groups
        weights.append([
            k - 1 if i < len(group) % num_groups else k
            for i in range(num_groups)
        ])
    return sum([
        weights[row][col]
        for row, col in Munkres().compute(weights)
    ])


def norm_distance[T](
      probe: CardSort[T],
      sort: CardSort[T],
      *,
      num_groups: int = None,
) -> float:
    """
    Computes the normalised edit distance between the given probe sort and
    card sort.

    If `num_groups` is not given, the number of groups will be unbounded.
    """
    if num_groups is None:
        num_groups = max(len(g) for g in probe)
    return distance(probe, sort) / max_distance(probe, num_groups=num_groups)



# Copyright 2025 Cardy Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Collection, Mapping, Callable
from itertools import combinations

from .distance import distance as edit_distance
from .types import CardSort

__all__ = ("DisjointSet", "min_spanning_tree", "orthogonality")


class DisjointSet[K]:
    def __init__(self, elements: Collection[K]):
        self.parents = {e: e for e in elements}
        self.sizes = {e: 1 for e in elements}

    def find(self, e: K) -> K:
        root = e
        while self.parents[root] != root:
            root = self.parents[root]
        while self.parents[e] != root:
            parent = self.parents[e]
            self.parents[e] = root
            e = parent
        return root

    def merge(self, e1: K, e2: K) -> None:
        e1 = self.find(e1)
        e2 = self.find(e2)
        if e1 == e2:
            return

        if self.sizes[e1] < self.sizes[e2]:
            e1, e2 = e2, e1
        self.parents[e2] = e1
        self.sizes[e1] += self.sizes[e2]


def min_spanning_tree[K](
      vertices: Collection[K],
      edges: Mapping[tuple[K, K], float]
) -> set[tuple[K, K]]:
    f = set()
    forest = DisjointSet(vertices)
    for (e1, e2), _ in sorted(edges.items(), key=lambda it: it[1]):
        if forest.find(e1) != forest.find(e2):
            f.add((e1, e2))
            forest.merge(e1, e2)
    return f


def orthogonality[T](
      sorts: Collection[CardSort[T]],
      *,
      distance: Callable[[T, T], float] = edit_distance,
) -> float:
    """
    Returns the orthogonality of the given collection of sorts.

    See: https://doi.org/10.1111/j.1468-0394.2005.00305.x

    :param sorts: A collection of sorts
    :param distance: A custom edit distance function
    :returns: the orthogonality of the sorts
    """
    sorts = {i: s for i, s in enumerate(sorts)}
    edges = {
        (e1, e2): distance(s1, s2)
        for (e1, s1), (e2, s2) in combinations(sorts.items(), 2)
    }
    tree = min_spanning_tree(sorts, edges)
    total_weight = sum(edges[e] for e in tree)
    return total_weight / len(sorts)



# Copyright 2025 Cardy Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Collection, Mapping, Callable
from random import Random

from .distance import distance as edit_distance
from .neighbourhood import neighbourhood
from .types import CardSort, CliqueHeuristic, Selector

__all__ = ("random_strategy", "greedy_strategy", "clique", "RandomSelector")


class RandomSelector:
    def __init__(self, seed=None):
        self.random = Random(seed)

    def select[T](self, collection: Collection[T]) -> T:
        """Selects an item from the given collection at random"""
        return self.random.sample(tuple(collection), k=1)[0]


def random_strategy[K, T](
      _: float,
      candidates: Mapping[K, CardSort[T]],
      *,
      selector: Selector = RandomSelector(),
      _distance: Callable[[T, T], float] = ...,
) -> K:
    """
    A heuristic strategy to select clique candidates at random.

    :param _: The max distance between any two sorts in the clique
    :param candidates: The intersection of the current clique
        sort neighbourhoods
    :param selector: An object to select an item from a collection at random
    :param _distance: An edit distance function
    :return: A randomly selected element
    """
    return selector.select(candidates)


def greedy_strategy[K, T](
      d: float,
      candidates: Mapping[K, CardSort[T]],
      *,
      selector: Selector = RandomSelector(),
      distance: Callable[[T, T], float] = edit_distance,
) -> K:
    """
    A heuristic strategy to select candidates from a set of sorts to add to a
    clique. See: <https://doi.org/10.1111/j.1468-0394.2005.00304.x>

    In the case where two or more candidates reduce the candidate pool by the
    same amount, one is chosen at random using the given selector.

    :param d: The max distance between any two sorts in the clique
    :param candidates: The intersection of the current clique
        sort neighbourhoods
    :param selector: An object to select an item from a collection at random
    :param distance: An edit distance function
    :return: An element that reduces the clique size by the smallest amount
    """
    current_max = 0
    max_candidates = []
    for key, candidate in candidates.items():
        size = len(neighbourhood(d, candidate, candidates, distance=distance))
        if size > current_max:
            max_candidates = [key]
            current_max = size
        elif size == current_max:
            max_candidates.append(key)
    return selector.select(max_candidates)


def clique[K, T](
      d: float,
      probe: CardSort[T],
      sorts: Mapping[K, CardSort[T]],
      *,
      selector: Selector[T] = RandomSelector(),
      strategy: CliqueHeuristic[K, T] = greedy_strategy,
      distance: Callable[[T, T], float] = edit_distance,
) -> set[K]:
    """
    Computes the d-clique centred around the given probe sort using the given
    heuristic strategy.

    The card sorts collection does not need to contain the probe sort. The probe
    sort will not be included in the result in this case.

    The strategy is a heuristic used to select candidate card sorts to add to
    the clique. See <https://doi.org/10.1111/j.1468-0394.2005.00304.x> for more.

    :param d: The max distance between any two sorts in the clique
    :param probe: The starting probe sort
    :param sorts: The collection of card sorts to search for the clique in
    :param selector: A selector
    :param strategy: The heuristic strategy for selecting candidates to add to
        the clique.
    :param distance: An edit distance function
    :return: A d-clique around the probe sort
    """
    clique_set = {
        key for key, sort in sorts.items() if distance(sort, probe) == 0
    }
    candidates = {
        key: sort for key, sort in sorts.items()
        if 0 < distance(sort, probe) <= d
    }
    while candidates:
        selected = strategy(d, candidates, selector=selector, distance=distance)
        clique_set.add(selected)
        candidates = {
            key: sort for key, sort in candidates.items()
            if key != selected and distance(sort, candidates[selected]) <= d
        }
    return clique_set



# Copyright 2025 Cardy Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable, Collection, Mapping, Set
from typing import Protocol

__all__ = ("CardSort", "CliqueHeuristic")


type CardSort[T] = Collection[Set[T]]


class Selector[T](Protocol):
    def select(self, collection: Collection[T]) -> T:
        ...



class CliqueHeuristic[K, T](Protocol):
    def __call__(
          self,
          d: float,
          candidates: Mapping[K, CardSort[T]],
          *,
          selector: Selector[T] = ...,
          distance: Callable[[T, T], float] = ...,
    ):
        ...



# Copyright 2025 Cardy Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Mapping, Callable

from .distance import distance as edit_distance
from .types import CardSort

__all__ = ("neighbourhood",)


def neighbourhood[K, T](
      d: float,
      probe: CardSort[T],
      sorts: Mapping[K, CardSort[T]],
      *,
      distance: Callable[[T, T], float] = edit_distance,
) -> set[K]:
    """
    Returns the d-neighbourhood of the given probe sort in the sorts iterable.

    The probe sort does not need to be one of the given sorts and will not be
    included in the result if it is not.

    :param d: The max distance neighbourhood elements and the probe
    :param probe: The sort at the centre of the neighbourhood
    :param sorts: A collection of sorts to search for the neighbourhood in
    :param distance: An edit distance function
    :return: The d-neighbourhood of the given probe
    """
    return {key for key, sort in sorts.items() if distance(probe, sort) <= d}



# Copyright 2025 Cardy Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import Collection

from cardy import clique, norm_distance
from cardy.clique import Selector, greedy_strategy
from utils import test


def load_examples():
    with open("examples.json", "r") as f:
        data = json.load(f)
    return {k: tuple(set(c) for c in v.values()) for k, v in data.items()}


class MinSelector(Selector):
    def select[T](self, collection: Collection[T]) -> T:
        return min(collection)


class MaxSelector(Selector):
    def select[T](self, collection: Collection[T]) -> T:
        return max(collection)


@test
def a_clique_of_equivalent_sorts_returns_the_whole_set_of_sort_ids():
    assert clique(0, ({1}, {2}), {0: ({1}, {2}), 1: ({2}, {1})}) == {0, 1}
    assert clique(1, ({1}, {2}), {0: ({1}, {2}), 1: ({2}, {1})}) == {0, 1}


@test
def cliques_do_not_necessarily_contain_the_probe_sort():
    assert clique(0, ({1, 2},), {0: ({1}, {2}), 1: ({2}, {1})}) == set()
    assert clique(1, ({1, 2},), {0: ({1}, {2}), 1: ({2}, {1})}) == {0, 1}


@test
def cliques_exclude_items_even_if_they_are_in_the_neighbourhood_of_the_card_sort():
    sorts = {
        0: ({1}, {2}, {3}),
        1: ({2, 3}, {1}),
        2: ({1, 2, 3},),
    }
    assert clique(1, ({1, 2}, {3}), sorts, selector=MinSelector()) == {0, 1}
    assert clique(1, ({1, 2}, {3}), sorts, selector=MaxSelector()) == {1, 2}


@test
def example_cliques():
    """
    Examples taken from: https://doi.org/10.1111/j.1468-0394.2005.00304.x
    """
    sorts = load_examples()
    # Table 5 Deibel et al.
    assert clique(4, sorts["table-5-0/8-4"], sorts) == {
        "table-5-0/8-4", "table-5-1", "table-5-2", "table-5-3", "table-5-4",
        "table-5-5", "table-5-6", "table-5-7", "table-5-8", "table-5-9/8-5",
    }
    # Table 6 Deibel et al.
    assert clique(
        5,
        sorts["table-6-0"],
        sorts,
        selector=MinSelector(),
    ) == {
        "table-6-0", "table-6-1", "table-6-2", "table-6-3",
        "table-6-4", "table-6-5", "table-6-6", "table-6-7",
    }
    # Table 7 Deibel et al.
    assert clique(5, sorts["table-7-0"], sorts) == {
        "table-7-0", "table-7-1", "table-7-2",
        "table-7-3", "table-7-4", "table-7-5",
    }
    # Table 8 Deibel et al.
    assert clique(
        5,
        sorts["table-8-0"],
        sorts,
        selector=MaxSelector(),
    ) == {
        "table-8-0", "table-8-1", "table-8-2", "table-8-3",
        "table-5-0/8-4", "table-5-9/8-5",
    }


@test
def example_norm_cliques():
    """
    Examples taken from: https://doi.org/10.1111/j.1468-0394.2005.00304.x
    """
    sorts = load_examples()
    dist = lambda l, r: norm_distance(l, r, num_groups=3)
    # Table 5 Deibel et al.
    assert clique(4 / 17, sorts["table-5-0/8-4"], sorts, distance=dist) == {
        "table-5-0/8-4", "table-5-1", "table-5-2", "table-5-3", "table-5-4",
        "table-5-5", "table-5-6", "table-5-7", "table-5-8", "table-5-9/8-5",
    }
    # Table 6 Deibel et al.
    assert clique(
        5 / 17,
        sorts["table-6-0"],
        sorts,
        selector=MinSelector(),
        distance=dist,
    ) == {
        "table-6-0", "table-6-1", "table-6-2", "table-6-3",
        "table-6-4", "table-6-5", "table-6-6", "table-6-7",
    }
    # Table 7 Deibel et al.
    assert clique(5 / 17, sorts["table-7-0"], sorts, distance=dist) == {
        "table-7-0", "table-7-1", "table-7-2",
        "table-7-3", "table-7-4", "table-7-5",
    }
    # Table 8 Deibel et al.
    assert clique(
        5 / 17,
        sorts["table-8-0"],
        sorts,
        selector=MaxSelector(),
        distance=dist,
    ) == {
        "table-8-0", "table-8-1", "table-8-2", "table-8-3",
        "table-5-0/8-4", "table-5-9/8-5",
    }
